# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from os.path import isfile
from metagenomix._io_utils import io_update


def get_cmd(
        self,
        focus_dir: str,
        db: str,
        analysis: str,
        select: set = None) -> str:
    """Build the command line for MIDAS analysis.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .soft.params
            Parameters
    focus_dir : str
        Path to the output folder.
    db : str
        Path to MIDAS database.
    analysis : str
        MIDAS analysis (any of "species", "genes" or "snps").
    select : set
        Species names for which there is a reference in the database.

    Returns
    -------
    cmd : str
        Midas command line for the species level.
    """
    cmd = 'run_midas.py %s' % analysis
    cmd += ' %s' % focus_dir
    cmd += ' -1 %s' % self.inputs[self.sam][0]
    if len(self.inputs[self.sam]) > 1:
        cmd += ' -2 %s' % self.inputs[self.sam][1]
    cmd += ' -d %s' % db
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' --remove_temp'
    if analysis != 'species':
        cmd += ' --species_cov 1'
    if select:
        cmd += ' --species_id %s' % ','.join(list(select))
    return cmd


def midas_species(self, focus_dir, db) -> None:
    if db == self.databases.paths['midas']:
        species_out = '%s/species' % focus_dir
        species_profile = '%s/species_profile.txt' % species_out
        if not self.config.force and isfile(species_profile):
            io_update(self, i_d=species_out)
        else:
            self.outputs['cmds'].append(get_cmd(self, focus_dir, db, 'species'))
            io_update(self, i_f=self.inputs[self.sam], o_d=species_out)
        self.outputs['outs'].append(species_out)
        self.outputs['dirs'].append(species_out)


def midas_genus(self, focus_dir, genes_out, db, select):
    if self.config.force or not isfile('%s/readme.txt' % genes_out):
        self.outputs['cmds'].append(
            get_cmd(self, focus_dir, db, 'genes', select))
        io_update(self, o_d=genes_out)
    self.outputs['outs'].append(genes_out)
    self.outputs['dirs'].append(genes_out)


def midas_snps(self, focus_dir, snps_out, db, select):
    if self.config.force or not isfile('%s/readme.txt' % snps_out):
        self.outputs['cmds'].append(
            get_cmd(self, focus_dir, db, 'snps', select))
        io_update(self, o_d=snps_out)
    self.outputs['outs'].append(snps_out)
    self.outputs['dirs'].append(snps_out)


def get_species_select(self, db: str, species_list: str) -> set:
    """Get the species names for which there is a reference in the database.

    Parameters
    ----------
    self : Commands class instance
        .config
            Configurations
    db : str
        Database name.
    species_list : str
        Path to file containing list of species to focus on.

    Returns
    -------
    select : set
        Species names for which there is a reference in the database.
    """
    select = set()
    if species_list:
        if self.config.dev and not isfile(species_list):
            select.add('Escherichia')
            return select
        else:
            species = [x.strip() for x in open(species_list).readlines()]
        with open('%s/species_info.txt' % db) as f:
            for line in f:
                for genus_species in species:
                    if genus_species.replace(' ', '_') in line:
                        select.add(line.split('\t')[0])
    return select


def midas(self) -> None:
    """Create command lines for MIDAS for the current database's species focus.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MIDAS
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .databases
            All databases class instance
        .config
            Configurations
    """
    for focus, (db, species_list) in self.soft.params['focus'].items():

        focus_dir = '%s/%s/%s' % (self.dir, focus, self.sam)
        midas_species(self, focus_dir, db)

        select = set(get_species_select(self, db, species_list))

        genes_out = '%s/genes' % focus_dir
        midas_genus(self, focus_dir, genes_out, db, select)

        snps_out = '%s/snps' % focus_dir
        midas_snps(self, focus_dir, snps_out, db, select)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import glob
import numpy as np
from os.path import isdir

from metagenomix._io_utils import read_yaml
from metagenomix.tools.alignment import *


def check_int(param: str, value, name):
    """Verifies that the value of the current parameter is an integer."""
    if not str(value).isdigit():
        raise sys.exit('Param "%s" for "%s" must be integer' % (param, name))


def check_mems(param, value, name):
    """Verifies that the requested "memory" is in kb, mb or gb."""
    mems = ['kb', 'mb', 'gb']
    if value not in mems:
        raise sys.exit('Param "%s" for "%s" must be of %s' % (
            param, name, str(mems)))


def check_env(config, value, name):
    """Verifies that the conda environment or module to load do exist."""
    if name not in config.modules and value not in config.conda_envs:
        sys.exit('"%s" not a module or conda env' % value)


def check_path(value, name):
    """Verifies that a path exists (usually for databases)."""
    if not isfile(value) and not isdir(value):
        sys.exit('path "%s" for "%s" do not exist' % (value, name))


def check_scratch(value, name):
    """Verifies that the scratch location is specified correctly."""
    if value not in ['scratch', 'userscratch'] and not isinstance(value, int):
        sys.exit('"%s" for "%s" must be "scratch", "userscratch" or an int' % (
            value, name))


def show_valid_params(param, values, name):
    """Verifies that the parameter value is authorized."""
    m = '[%s] user parameter "%s" must be among:\n' % (name, param)
    for value in values:
        m += '  - %s\n' % value
    sys.exit(m)


def check_nums(params, defaults, vals, dtype, tool, mi=None, ma=None):
    for param in vals:
        if param not in params:
            params[param] = defaults[param]
        else:
            if dtype == int:
                if not isinstance(params[param], dtype):
                    sys.exit('[%s] Param "%s" not of %s' % (tool, param, dtype))
            else:
                try:
                    float(params[param])
                except ValueError:
                    sys.exit('[%s] Param "%s" not a number' % (tool, param))
        if mi and ma and not mi <= dtype(params[param]) <= ma:
            sys.exit('[%s] Param "%s" not in [%s-%s]' % (tool, param, mi, ma))


def check_databases(name, params, databases):
    """Verifies that the databases are set in the user parameters.
        - If only one database exists for a given list of databases,
    then the non-existing databases are ignored and the tool will use
    only this one database.
        - If no database is found to exist at all, then metagenomix
    will exit with a helpful message."""
    if 'databases' not in params:
        sys.exit('[%s] "databases" must be a parameter' % name)
    dbs_existing = []
    dbs_missing = []
    for db in params['databases']:
        if db in databases.paths or db == 'default':
            dbs_existing.append(db)
        else:
            dbs_missing.append(db)
    if not dbs_existing:
        sys.exit('[%s] No databases: %s' % (name, '; '.join(dbs_missing)))
    elif dbs_missing:
        print('[%s] Missing databases: "%s"' % (name, '", "'.join(dbs_missing)))
    return dbs_existing


def check_default(params, defaults, name, let_go: list = []):
    """Verifies that the parameters given by the used for the current tool
    and that are to be checked (i.e., not in the `let_go` list) are valid.
    This mean that they must be one (or sevreral) of the values matching the
    passed defaults. If a default parameter is not given by the user,
    then the parameter is created and set to the first as value of the
    passed defaults.
    Note: a list of parameters for which the value has to be a list (such as
          "aligners" or "models") will have the default value set to a list.
    """
    for param, values in defaults.items():
        if param in let_go:  # skip params checked specifically per software
            continue
        if param not in params:
            if param in ['aligners', 'blobology', 'reassembly', 'models']:
                params[param] = [values[0]]
            else:
                params[param] = values[0]
        else:
            if isinstance(params[param], list):
                if set(sorted(params[param])).difference(values):
                    show_valid_params(param, values, name)
            elif params[param] not in values:
                show_valid_params(param, values, name)


class Parameters(object):
    """
    This class shall contains the checks for all tools specifically,
    but since these rely on the systematic elaboration of `defaults` and
    `dtypes` data structures that are useful later, it is best to consign
    these inside class instances.
    """

    def __init__(self, soft):
        self.name = soft.name
        self.defaults = {}
        self.dtypes = {}
        self.let_go = {}
        self.params = {}

    def check_default(self):
        """Verifies that the parameters given by the used for the current tool
        and that are to be checked (i.e., not in the `let_go` list) are valid.
        This mean that they must be one (or sevreral) of the values matching the
        passed defaults. If a default parameter is not given by the user,
        then the parameter is created and set to the first as value of the
        passed defaults.
        Note: a list of parameters for which the value has to be a list (such as
              "aligners" or "models") will have the default value set to a list.
        """
        for param, values in self.defaults.items():
            if param in self.let_go:  # skip params checked per software
                continue
            if param not in self.params:
                if param in ['aligners', 'blobology', 'reassembly',
                             'models']:
                    self.params[param] = [values[0]]
                else:
                    self.params[param] = values[0]
            else:
                if isinstance(self.params[param], list):
                    if set(sorted(self.params[param])).difference(values):
                        self.show_valid_params(param, values)
                elif self.params[param] not in values:
                    self.show_valid_params(param, values)

    def show_valid_params(self, param, values):
        """Verifies that the parameter value is authorized."""
        m = '[%s] user parameter "%s" must be among:\n' % (self.name, param)
        for value in values:
            m += '  - %s\n' % value
        sys.exit(m)

    def check_nums(self, vals, dtype, tool, mi=None, ma=None):
        for param in vals:
            if param not in self.params:
                self.params[param] = self.defaults[param]
            else:
                if dtype == int:
                    if not isinstance(self.params[param], dtype):
                        sys.exit('[%s] Param "%s" not of %s' % (
                            tool, param, dtype))
                else:
                    try:
                        float(self.params[param])
                    except ValueError:
                        sys.exit('[%s] Param "%s" not a number' % (
                            tool, param))
            if mi and ma and not mi <= dtype(self.params[param]) <= ma:
                sys.exit('[%s] Param "%s" not in [%s-%s]' % (
                    tool, param, mi, ma))

    def check_databases(self, databases):
        """Verifies that the databases are set in the user parameters.
            - If only one database exists for a given list of databases,
        then the non-existing databases are ignored and the tool will use
        only this one database.
            - If no database is found to exist at all, then metagenomix
        will exit with a helpful message."""
        if 'databases' not in self.params:
            sys.exit('[%s] "databases" must be a parameter' % self.name)
        dbs_existing = []
        dbs_missing = []
        for db in self.params['databases']:
            if db in databases.paths or db == 'default':
                dbs_existing.append(db)
            else:
                dbs_missing.append(db)
        placeholders = (self.name, '; '.join(dbs_missing))
        if not dbs_existing:
            sys.exit('[%s] No databases: %s' % placeholders)
        elif dbs_missing:
            print('[%s] Missing databases: "%s"' % placeholders)
        return dbs_existing


# ============================================= #
#  Below are the function that are called from  #
#  class method Workflow.set_params()           #
#       (see module metagenomix.pipeline)       #
# ============================================= #


def check_search(self, params, soft):
    tool = soft.name.rsplit('_')[1]
    defaults = {}
    if tool == 'diamond':
        defaults.update(search_diamond(params, soft))
        dbs_existing = check_databases(tool, params, self.databases)
        valid_dbs = {}
        for db in dbs_existing:
            path = self.databases.paths[db]
            dmnds = '%s/diamond/*.dmnd' % path
            if not self.config.dev:
                dmnd_paths = glob.glob(dmnds)
                if dmnd_paths:
                    valid_dbs[db] = dmnd_paths
            else:
                valid_dbs[db] = [dmnds]
        params['databases'] = valid_dbs
        print(valid_dbs)

    if tool == 'hmmer':
        defaults.update(search_hmmer(params, soft))
        if 'terms' not in params:
            print('[search] Params "%s:terms" missing (search ignored)' % tool)
        else:
            terms = str(params['terms'])
            self.databases.get_pfam_terms(params)
            if not params['terms']:
                print('[search] Params "%s:terms" has no HMM profile for %s' % (
                    tool, terms))
    defaults['databases'] = '<list of databases>'
    return defaults


def check_ccmap(self, params, soft):
    defaults = {
        'preserve_tmpdir': [False, True], 'terminal_fragment_size': 500,
        'min_percent_identity': 94, 'min_aligned_length': 50}
    check_nums(params, defaults,
               ['terminal_fragment_size', 'min_aligned_length'], int, soft.name)
    check_nums(params, defaults, ['min_percent_identity'],
               int, soft.name, 0, 100)
    let_go = [
        'terminal_fragment_size', 'min_aligned_length', 'min_percent_identity']
    check_default(params, defaults, soft.name, let_go)
    return defaults


def check_barrnap(self, params, soft):
    defaults = {
        'evalue': 1e-06, 'reject': 0.25, 'lencutoff': 0.8,
        'incseq': [False, True], 'kingdom': ['bac', 'mito', 'arc', 'euk']}
    floats = ['evalue', 'reject', 'lencutoff']
    check_nums(params, defaults, floats, float, soft.name, 0, 1)
    check_default(params, defaults, soft.name, floats)
    return defaults


def check_integron_finder(self, params, soft):
    defaults = {
        'evalue_attc': 1, 'min_length': 1500,
        'min_attc_size': 40, 'max_attc_size': 200,
        'pdf': [True, False], 'gbk': [True, False], 'mute': [True, False],
        'local_max': [True, False], 'promoter_attI': [True, False],
        'union_integrases': [False, True], 'topology': ['linear', 'circ'],
    }
    check_nums(params, defaults, ['min_length', 'min_attc_size',
                                  'max_attc_size'], int, soft.name)
    check_nums(params, defaults, ['evalue_attc'], float, soft.name, 0, 100)
    let_go = ['min_length', 'min_attc_size', 'max_attc_size', 'evalue_attc']
    check_default(params, defaults, soft.name, let_go)
    defaults['prot_file'] = '<Path to proteins file used for annotations>'
    defaults['attc_model'] = '<Path to the attc model (Covariance Matrix)>'
    defaults['topology_file'] = '<Path to file with each replicon topology>'
    return defaults


def check_prokka_config(params):
    t = params['config']
    cols = ['genus', 'species', 'strain', 'plasmid', 'proteins']
    with open(t) as f:
        for ldx, line in enumerate(f):
            if ldx == 10:  # only check 10 rows
                break
            ls = line.strip('\n').split('\t')
            if not ldx:
                if ls != cols:
                    sys.exit('[prokka] Wrong headers in "config" param file '
                             '("%s"): must be "%s"' % (t, '\\t'.join(cols)))
            if len(ls) != 5:
                sys.exit('[prokka] "config" param file ("%s") do not have 5 '
                         'tab-separated fields (could be empty):\t%s' % (t, ls))


def check_prokka(self, params, soft):
    defaults = {
        'kingdom': ['Archaea', 'Bacteria', 'Mitochondria', 'Viruses'],
        'evalue': 1e-09, 'coverage': 80, 'mincontiglen': 100,
        'gcode': [11, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        'notrna': [False, True], 'norrna': [False, True],
        'metagenome': [True, False]
    }
    check_nums(params, defaults, ['evalue'], float, soft.name, 0, 100)
    check_nums(params, defaults, ['coverage'], int, soft.name, 0, 100)
    check_nums(params, defaults, ['mincontiglen'], int, soft.name)
    check_default(params, defaults, soft.name, ['evalue', 'coverage',
                                                'mincontiglen'])
    if 'config' in params:
        if not isfile(params['config']):
            sys.exit('[prokka] Param "config" must be an existing file')
        check_prokka_config(params)
    else:
        params['config'] = None
    defaults['config'] = '<Path to config file for prokka>'
    return defaults


def check_antismash(self, params, soft):
    defaults = {
        'tta-threshold': 0.65,
        'rre': [True, False],
        'asf': [True, False],
        'cassis': [True, False],
        'pfam2go': [True, False],
        'tigrfam': [True, False],
        'cc-mibig': [True, False],
        'fullhmmer': [True, False],
        'cb_general': [True, False],
        'smcog_trees': [True, False],
        'clusterhmmer': [True, False],
        'cb_subclusters': [True, False],
        'cb_knownclusters': [True, False],
        'taxon': ['bacteria', 'fungi'],
        'genefinding_tool': [
            'prodigal-m', 'prodigal', 'glimmerhmm', 'none', 'error']
    }
    check_nums(params, defaults, ['tta-threshold'], float, soft.name, 0, 1)
    check_default(params, defaults, soft.name, ['tta-threshold'])
    return defaults


def check_quast(self, params, soft):
    defaults = {
        'min_contig': 1000,
        'min_alignment': 65,
        'min_identity': 95.0,
        'circos': [False, True],
        'glimmer': [False, True],
        'rna_finding': [False, True],
        'conserved_genes_finding': [False, True],
        'space_efficient': [False, True],
        'ambiguity_usage': ['all', 'none', 'one'],
        'fast': [True, False],  # activates all below
        'no_check': [True, False],
        'no_plots': [True, False],
        'no_html': [True, False],
        'no_icarus': [True, False],
        'no_snps': [True, False],
        'no_gc': [True, False],
        'no_sv': [True, False],
        'no_read_stats': [True, False],
    }
    ints = ['min_contig', 'min_alignment']
    floats = ['min_identity']
    check_nums(params, defaults, ints, int, soft.name)
    check_nums(params, defaults, floats, float, soft.name, 0, 1)
    if 'label' in params:
        label = params['label']
        if label not in set(self.config.meta):
            sys.exit('[quast] Params "label" must be a valid metadata column')
    check_default(params, defaults, soft.name, (ints + floats + ['label']))
    defaults['label'] = '<an existing metadata column>'
    return defaults


def check_filtering(self, params, soft):
    defaults = {'databases': None}
    if 'databases' not in params or not isinstance(params['databases'], list):
        sys.exit('[filtering] Param "databases" must a list of bowtie2 dbs')
    else:
        for d in params['databases']:
            if not self.config.dev and not glob.glob('%s/*' % d):
                sys.exit('[filtering] Param "databases" no "%s" bowtie2 db' % d)
    defaults['databases'] = '<list of databases>'
    return defaults


def check_atropos(self, params, soft):
    defaults = {
        'q': 15,
        'overlap': 3,
        'max_reads': 0,
        'indel_cost': 1,
        'error_rate': 0.1,
        'nextseq_trim': 5,
        'minimum_length': 50,
        'quality_cutoff': 15,
        'pair_filter': ['any', 'both'],
        'aligner': ['adapter', 'insert'],
        'stats': ['none', 'pre', 'post', 'both'],
        'report_formats': ['txt', 'json', 'yaml', 'pickle'],
        'a': ['GATCGGAAGAGCACACGTCTGAACTCCAGTCAC', 'GGGGGGGGGG'],
        'A': ['AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGT', 'GGGGGGGGGG']
    }
    ints = ['q', 'overlap', 'max_reads', 'indel_cost', 'nextseq_trim',
            'minimum_length', 'quality_cutoff']
    floats = ['error_rate']
    check_nums(params, defaults, ints, int, soft.name)
    check_nums(params, defaults, floats, float, soft.name, 0, 1)
    check_default(params, defaults, soft.name, (floats + ints + ['a', 'A']))
    for aA in ['a', 'A']:
        if aA not in params:
            params[aA] = []
        if not isinstance(params[aA], list):
            sys.exit('[atropos] Param "%s" must be a list' % aA)
    return defaults


def check_kneaddata(self, params, soft):
    defaults = {'trimmomatic': None, 'bowtie2': None, 'databases': [],
                'purge': [True, False]}
    tools = ['trimmomatic', 'bowtie2']
    for tool in tools:
        if tool not in params:
            params[tool] = defaults[tool]
        elif not self.config.dev and not glob.glob('%s*' % params[tool]):
            print('[kneaddata] Param "%s": no path found (use $PATH)' % tool)
    if 'databases' not in params or not isinstance(params['databases'], list):
        sys.exit('[kneaddata] Params "databases" must a list of existing paths')
    check_default(params, defaults, soft.name, (tools + ['databases']))
    defaults['databases'] = '<list of databases>'
    defaults['bowtie2'] = '<path to bowtie2 (default to $PATH)>'
    defaults['trimmomatic'] = '<path to trimmomatic (default to $PATH)>'
    return defaults


def del_search_tool_keys(params):
    for name in ['diamond', 'hmmer']:
        if name in params:
            del params[name]


def expand_search_params(params, defaults, name):
    if name not in params:
        for k, v in defaults.items():
            if isinstance(v, list):
                params[k] = v[0]
            else:
                params[k] = v
    else:
        for k, v in params[name].items():
            params[k] = v
    del_search_tool_keys(params)


def search_diamond(params, soft) -> dict:
    defaults = {'mode': [False, 'fast', 'mid-sensitive', 'sensitive',
                         'more-sensitive', 'very-sensitive', 'ultra-sensitive'],
                'k': 1, 'top': 2, 'evalue': 0.001,
                'identity': 80, 'query_cov': 80}
    expand_search_params(params, defaults, 'diamond')
    ints, evals, ktop = ['identity', 'query_cov'], ['evalue'], ['k', 'top']
    check_nums(params, defaults, ints, int, soft.name, 0, 100)
    check_nums(params, defaults, evals, float, soft.name, 0, 1)
    check_nums(params, defaults, ktop, int, soft.name)
    check_default(params, defaults, soft.name, (ints + evals + ktop))
    return defaults


def search_hmmer(params, soft) -> dict:
    defaults = {'max': [True, False], 'nobias': [True, False],
                'Z': 10, 'E': 1, 'domZ': 10, 'domE': 1}
    expand_search_params(params, defaults, 'hmmer')
    e_vals, z_vals = ['E', 'domE'], ['Z', 'domZ']
    check_nums(params, defaults, e_vals, float, soft.name, 0, 100)
    check_nums(params, defaults, z_vals, int, soft.name, 0, 1000)
    check_default(params, defaults, soft.name, (e_vals + z_vals + ['terms']))
    return defaults


def check_count(self, params, soft):
    defaults = {'cat': ['zcat', 'gzcat']}
    check_default(params, defaults, soft.name)
    return defaults


def check_kraken2(self, params, soft):
    defaults = {'databases': ['default'], 'confidence': 0.5}
    check_databases('bowtie2', params, self.databases)
    check_nums(params, defaults, ['confidence'], float, soft.name, 0, 1)
    check_default(params, defaults, soft.name, ['confidence', 'databases'])
    defaults['databases'].append('<list of databases>')
    return defaults


def check_shogun(self, params, soft):
    defaults = {'aligners': ['bowtie2', 'burst', 'utree']}
    check_default(params, defaults, soft.name)
    if 1:
        valid_dbs = {}
        dbs_existing = check_databases('shogun', params, self.databases)
        for db in dbs_existing:
            path = self.databases.paths[db]
            metadata_yml = '%s/shogun/metadata.yaml' % path
            if not self.config.dev and not isfile(metadata_yml):
                raise IOError('[shogun] file "%s" must exist' % metadata_yml)
            metadata = read_yaml(metadata_yml)
            for aligner in list(params['aligners']):
                if aligner in metadata:
                    if not self.config.dev:
                        ali = metadata[aligner]
                        if ali[0] == '/':
                            formatted = ali
                        else:
                            formatted = '%s/shogun/%s' % (path, ali)
                        if glob.glob('%s*' % formatted):
                            valid_dbs.setdefault(db, []).append(aligner)
                        else:
                            print('[shogun] No formatted db "%s"' % formatted)
                    else:
                        valid_dbs.setdefault(db, []).append(aligner)
        params['databases'] = valid_dbs
        if not params['databases']:
            print('[shogun] No database formatted for shogun: will be skipped')
    defaults['databases'] = '<list of databases>'
    return defaults


def check_bowtie2(self, params, soft):
    defaults = {
        # default: 'paired'
        'pairing': ['paired', 'concat', 'single'],
        'discordant': [True, False],
        'k': '16', 'np': '1',
        'mp': '1,1', 'rdg': '0,1', 'rfg': '0,1',
        'score-min': 'L,0,-0.05'
    }
    let_go = []
    check_bowtie_k_np(soft, params, defaults, let_go)
    check_bowtie_mp_rdg_rfg(soft, params, defaults, let_go)
    check_bowtie_score_min(soft, params, defaults, let_go)
    check_default(params, defaults, soft.name, let_go)
    dbs_existing = check_databases('bowtie2', params, self.databases)
    valid_dbs = {}
    for db in dbs_existing:
        path = self.databases.paths[db]
        bt2_path = '%s/bowtie2/*.*.bt2' % path
        if not self.config.dev:
            bt2_paths = glob.glob(bt2_path)
            if bt2_paths:
                valid_dbs[db] = bt2_paths[0].rsplit('.', 2)[0]
        else:
            valid_dbs[db] = bt2_path.rsplit('.', 2)[0]
    params['databases'] = valid_dbs
    defaults['databases'] = '<list of databases>'
    return defaults


def check_spades(self, params, soft):
    defaults = {'k': ['33', '55', '77', '99', '127'],
                'bio': [False, True],
                'meta': [True, False],
                'plasmid': [False, True],
                'only_assembler': [True, False]}
    if 'k' not in params:
        params['k'] = defaults['k']
    else:
        kerrors = [x for x in params['k'] if not str(x).isdigit()]
        if len(kerrors):
            sys.exit('[spades] "k" must be integers (%s)' % ','.join(kerrors))
    check_default(params, defaults, soft.name, ['k'])
    return defaults


def check_viralverify(self, params, soft):
    defaults = {'thr': 7, 'p': [False, True], 'db': [False, True]}
    check_nums(params, defaults, ['thr'], int, 'viralverify')
    check_default(params, defaults, soft.name, ['thr'])
    return defaults


def simka_error(key: str, kmer_reads: dict) -> None:
    """Throw an error is the user-defined simka params are not right.

    Parameters
    ----------
    key : str
        "" or ""
    kmer_reads : dict
        Params for sampling of the kmers of reads size in Simka
    """
    for p in ['start', 'end', 'size']:
        if p not in kmer_reads or not isinstance(kmer_reads[p], int):
            sys.exit('Please give an integer for simka subparam "%s" (param '
                     '"%s") in your input .yml params file' % (p, key))


def check_simka(self, params, soft):
    defaults = {'simkaMin': [True, False],
                'kmer': np.linspace(15, 80, 6),
                'log_reads': np.logspace(3, 7, 3)}
    if 'kmer' not in params:
        params['kmer'] = defaults['kmer']
    else:
        kmer = params['kmer']
        simka_error('kmer', kmer)
        params['kmer'] = np.linspace(kmer['start'], kmer['end'], kmer['size'])
    if 'log_reads' not in params:
        params['log_reads'] = defaults['log_reads']
    else:
        n = params['log_reads']
        simka_error('log_reads', n)
        params['log_reads'] = np.logspace(n['start'], n['end'], n['size'])
    check_default(params, defaults, soft.name, ['kmer', 'log_reads'])
    params['kmer'] = [int(x) for x in params['kmer']]
    params['log_reads'] = [int(x) for x in params['log_reads']]
    defaults['path'] = '<Path to the SimKa program folder>'
    defaults['kmer'] = [str(x) for x in np.linspace(15, 80, 6)]
    defaults['log_reads'] = [str(x) for x in np.logspace(3, 7, 3)]
    return defaults


def check_metamarker(self, params, soft):
    defaults = {'identity': 0.9}
    check_nums(params, defaults, ['identity'], float, 'metamarker', 0, 1)
    if 'groups' not in params:
        sys.exit('[metamarker] Param "groups" not found (metadata variable(s))')
    for group in params['groups']:
        if group not in set(self.config.meta.columns):
            sys.exit('[metamarker] metadata variable "%s" not found' % group)
    defaults['groups'] = '<List of metadata columns>'
    return defaults


def check_metawrap(self, params, soft):
    defaults = {
        'binners': ['maxbin2', 'metabat2', 'concoct'],
        'min_completion': 25, 'min_contamination': 5, 'min_contig_length': 1000,
        'min_completion_reassembly': 25, 'min_contamination_reassembly': 5,
        'reassembly': ['permissive', 'strict'],
        'blobology': ['coassembly', 'sample']
    }
    if 'binners' not in params:
        params['binners'] = defaults['binners']
    mins = ['min_completion', 'min_completion_reassembly',
            'min_contamination', 'min_contamination_reassembly']
    check_nums(params, defaults, mins, int, 'metawrap:binning', 0, 100)
    check_nums(params, defaults, ['min_contig_length'], int, 'metawrap:binning')
    mins.append('min_contig_length')
    check_default(params, defaults, soft.name, mins)
    return defaults


def check_flash(self, params, soft):
    defaults = {'min_overlap': 30, 'max_overlap': 100, 'mismatch': 0}
    check_nums(params, defaults, list(defaults.keys()), int, soft.name)
    mi, ma = params['min_overlap'], params['max_overlap']
    if mi > ma:
        sys.exit('[flash] "min_overlap" (%s) > "max_overlap" (%s)' % (mi, ma))
    return defaults


def check_prodigal(self, params, soft):
    defaults = {'procedure': ['meta', 'single']}
    check_default(params, defaults, soft.name)
    return defaults


def check_fastp(self, params, soft):
    defaults = {'average_qual': 20,
                'trim_poly_g': [True, False]}
    check_nums(params, defaults, ['average_qual'], int, 'fastp')
    check_default(params, defaults, soft.name, ['average_qual'])
    return defaults


def check_humann(self, params, soft):
    defaults = dict({
        'evalue': 0.1,
        'identity': 90,
        'query_cov': 80,
        'subject_cov': 80,
        'prescreen_threshold': 0.001,
        'uniref': ['uniref90'],
        'skip_translated': [False, True],
        'nucleotide_db': '/path/to/full_chocophlan.v0.1.1/chocophlan',
        'protein_db': '/path/to/uniref90/uniref',
    })
    let_go = ['profiles']
    param_dtype = [('evalue', float), ('prescreen_threshold', float),
                   ('identity', int), ('query_cov', int), ('subject_cov', int)]
    for (param, dtype) in param_dtype:
        let_go.append(param)
        if param not in params:
            params[param] = defaults[param]
        else:
            if not isinstance(params[param], dtype):
                sys.exit('[humann] Param "%s" not an int' % param)
        if not 0 <= float(params[param]) <= 100:
            sys.exit('[humann] Param "%s" not in [0-100]' % param)

    if not self.config.dev:
        if 'nucleotide_db' not in params or not isdir(params['nucleotide_db']):
            sys.exit('[humann] Param "nucleotide_db" must be an existing path')
        if 'protein_db' not in params or not isdir(params['protein_db']):
            sys.exit('[humann] Param "protein_db" must be an existing path')
    if 'profiles' in params:
        if not isinstance(params['profiles'], dict):
            sys.exit('[humann] Param "profiles" must be a dict structure')
        for key, value in params['profiles'].items():
            if self.config.dev:
                continue
            if not isinstance(value, str):
                sys.exit('[humann] Param "profiles::%s" must be a string' % key)
            elif not isfile(value):
                sys.exit('[humann] Param "profiles::%s" does not exist' % key)
    check_default(params, defaults, soft.name, let_go)
    defaults['profiles'] = '<dict of taxonomic profiles name: file>'
    return defaults


def check_midas(self, params, soft):
    defaults = {'focus': {'all': [self.databases.paths['midas'], '']}}
    if 'tracking' not in params:
        params['tracking'] = []
    else:
        if not isinstance(params['tracking'], list):
            sys.exit('[midas] Param "tracking" must be a list structure')
        meta_vars = set(self.config.meta.columns)
        params['tracking'] = [x for x in params['tracking'] if x in meta_vars]
        v = [x for x in params['tracking'] if x not in meta_vars]
        if len(v):
            print('[midas] Param "tracking" no metadata var "%s" (ignored)' % v)
        if not params['tracking']:
            print('[midas] Param "tracking" no metadata variable (ignored)')
    if 'focus' not in params:
        params['focus'] = defaults['focus']
    else:
        if not isinstance(params['focus'], dict):
            sys.exit('[midas] Param "focus" must be a dict structure')
        for k, v in params['focus'].items():
            if not isinstance(v, list) or len(v) != 2:
                sys.exit('[midas] Param "focus::%s" must be 2-items list' % k)
            if not self.config.dev and not isfile(v[1]):
                sys.exit('[midas] Param "focus::%s::%s" not a file' % (k, v[1]))
    defaults = {'focus': '<dict of >'}
    return defaults


def check_macsyfinder(self, params, soft):
    defaults = {'db_type': (['unordered', 'ordered_replicon', 'gembase'], str),
                'replicon_topology': (['linear', 'circular'], str),
                'models': (['TXSS', 'TFF-SF', 'CAS'], list),
                'evalue': (0.1, float),
                'coverage': (0.5, float)}
    check_nums(params, defaults, ['evalue', 'coverage'], float, soft.name, 0, 1)
    let_go = ['evalue', 'coverage']
    check_default(params, defaults, soft.name, let_go)
    return defaults


# def check_TOOL(params, soft, databases, config):
#     defaults = {}



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
import glob
from os.path import isdir, isfile

from metagenomix._cmds import caller
from metagenomix._io_utils import io_update


def get_sams_fastqs(mode: str, fastqs: dict):
    """Get the paths of the fastq files per sample or altogether.

    Parameters
    ----------
    mode : str
        Mode of processing for blobology:
        - "coassembly": all co-assembled samples.
        - "sample": each sample independently.
    fastqs : dict
        Path to the raw fastq files per sample.

    Returns
    -------
    sams_fastqs : dict
        Either same as `fastqs` of all fastqs under an empty key.
    """
    sams_fastqs = fastqs
    if mode == 'coassembly':
        sams_fastqs = {'': [y for x in fastqs.values() for y in x]}
    return sams_fastqs


def quantify(self):
    for group in self.pools[self.pool]:

        bins = self.inputs[self.pool][group][-1]
        contigs = self.softs['spades'].outputs[self.pool][group][1]
        fastqs = {sam: self.config.fastq[sam] for sam
                  in self.pools[self.pool][group]}

        self.outputs['outs'][group] = {}
        io_update(self, i_f=contigs, i_d=bins, key=group)

        for mode in self.soft.params['blobology']:
            self.outputs['outs'][group][mode] = {}

            out = '%s/%s/%s/%s' % (self.dir, self.pool, group, mode)
            sams_fastqs = get_sams_fastqs(mode, fastqs)
            for sam, fastq in sams_fastqs.items():
                if sam:
                    out += '/%s' % sam
                cmd = 'metawrap quant_bins'
                cmd += ' -b %s' % bins
                cmd += ' -o %s' % out
                cmd += ' -a %s' % contigs
                cmd += ' -t %s' % self.soft.params['cpus']
                cmd += ' %s' % ' '.join(fastq)

                self.outputs['dirs'].append(out)
                self.outputs['outs'][group][mode][sam] = out
                self.outputs['cmds'].setdefault(group, []).append(cmd)
                io_update(self, i_f=fastq, o_d=out, key=group)


def classify_or_annotate(self, command):
    for group in self.pools[self.pool]:

        if self.soft.prev == 'metawrap_refine':
            bin_dirs = {'': [self.inputs[self.pool][group][1]]}
        elif self.soft.prev == 'metawrap_reassemble':
            bin_dirs = self.inputs[self.pool][group]
        else:
            sys.exit('No metawrap classify_bins after %s' % self.soft.prev)

        self.outputs['outs'][group] = {}

        for sam, bin_dir in bin_dirs.items():
            for bin in bin_dir:
                out = '%s/%s/%s' % (self.dir, self.pool, group)
                if 'permissive' in bin:
                    out += '/permissive'
                elif 'strict' in bin:
                    out += '/strict'
                if sam:
                    out += '/%s' % sam
                cmd = 'metawrap %s' % command
                cmd += ' -b %s' % bin
                cmd += ' -o %s' % out
                cmd += ' -t %s' % self.soft.params['cpus']

                self.outputs['dirs'].append(out)
                self.outputs['outs'][group].setdefault(sam, []).append(out)
                io_update(self, i_d=bin, o_d=out, key=group)

                if command == 'annotate_bins':
                    tab = glob.glob('%s/bin_funct_annotations/*.tab' % out)
                    if self.config.force or not tab:
                        self.outputs['cmds'].setdefault(group, []).append(cmd)
                elif command == 'classify_bins':
                    tab = '%s/bin_taxonomy.tab' % out
                    if self.config.force or not isfile(tab):
                        self.outputs['cmds'].setdefault(group, []).append(cmd)


def classify(self):
    classify_or_annotate(self, 'classify_bins')


def annotate(self):
    classify_or_annotate(self, 'annotate_bins')


def blobology(self):
    for group in self.pools[self.pool]:
        self.outputs['outs'][group] = {}
        bins = self.inputs[self.pool][group][-1]
        contigs = self.softs['spades'].outputs[self.pool][group][1]
        fastqs = {
            sam: self.config.fastq[sam] for sam in self.pools[self.pool][group]}
        io_update(self, i_f=contigs, i_d=bins, key=group)
        for mode in self.soft.params['blobology']:
            self.outputs['outs'][group][mode] = {}
            out = '%s/%s/%s/%s' % (self.dir, self.pool, group, mode)
            sams_fastqs = get_sams_fastqs(mode, fastqs)
            for sam, fastq in sams_fastqs.items():
                if sam:
                    out += '/%s' % sam
                self.outputs['dirs'].append(out)
                io_update(self, i_f=fastq, o_d=out, key=group)
                plot = '%s/contigs.binned.blobplot' % out
                if self.config.force or not isfile(plot):
                    cmd = 'metawrap blobology'
                    cmd += ' -o %s' % out
                    cmd += ' -a %s' % contigs
                    cmd += ' -t %s' % self.soft.params['cpus']
                    cmd += ' --bins %s' % bins
                    cmd += ' %s' % ' '.join(fastq)
                    self.outputs['cmds'].setdefault(group, []).append(cmd)
                    self.outputs['outs'][group][mode][sam] = out


def reassembly_bins_cmd(self, sam, out, bins):
    cmd = 'metawrap reassemble_bins'
    cmd += ' -o %s' % out
    cmd += ' -1 %s' % self.config.fastq[sam][0]
    cmd += ' -2 %s' % self.config.fastq[sam][1]
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' -m %s' % self.soft.params['mem_num']
    cmd += ' -c %s' % self.soft.params['min_completion_reassembly']
    cmd += ' -x %s' % self.soft.params['min_contamination_reassembly']
    cmd += ' -b %s' % bins
    cmd += ' --skip-checkm'
    if self.soft.params['cpus'] > 1:
        cmd += ' --parallel\n'
    return cmd


def reassembly_cmd(self, bins, group, sam, out):
    cmd = ''
    for mode in self.soft.params['reassembly']:
        mode_dir = '%s/reassembled_bins_%s' % (out, mode)
        io_update(self, i_d=mode_dir, key=group)
        if not self.config.force and isdir(mode_dir):
            continue
        cmd += 'mkdir %s\n' % mode_dir
        cmd += 'mv %s/reassembled_bins/*%s.fa %s/.\n' % (out, mode, mode_dir)
    if cmd:
        cmd = reassembly_bins_cmd(self, sam, out, bins) + cmd
    return cmd


def reassemble(self):
    for group in self.pools[self.pool]:
        bins = self.inputs[self.pool][group][-1]
        io_update(self, i_d=bins, key=group)
        for sam in self.pools[self.pool][group]:
            out = '%s/%s/%s/%s' % (self.dir, self.pool, group, sam)
            self.outputs['outs'].setdefault(group, []).append(out)
            self.outputs['dirs'].append(out)
            io_update(self, i_f=self.config.fastq[sam], o_d=out, key=group)
            cmd = reassembly_cmd(self, bins, group, sam, out)
            if cmd:
                self.outputs['cmds'].setdefault(group, []).append(cmd)


def refine_cmd(self, out_dir, bin_folders) -> tuple:
    n_bins = 0
    cmd = 'metawrap bin_refinement'
    cmd += ' -o %s' % out_dir
    for fdx, folder in enumerate(bin_folders):
        if len(glob.glob('%s/bin.*.fa' % folder)) or self.config.force:
            n_bins += 1
        cmd += ' -%s %s' % (['A', 'B', 'C'][fdx], folder)
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' -c %s' % self.soft.params['min_completion']
    cmd += ' -x %s' % self.soft.params['min_contamination']
    return cmd, n_bins


def refine(self):
    for group in self.pools[self.pool]:
        out_dir = '%s/%s/%s' % (self.dir, self.pool, group)
        bin_folders = self.inputs[self.pool][group][:-1]

        cmd, n_bins = refine_cmd(self, out_dir, bin_folders)
        if n_bins == len(bin_folders):
            self.outputs['cmds'].setdefault(group, []).append(cmd)
        out = '%s/metawrap_%s_%s' % (out_dir,
                                     self.soft.params['min_completion'],
                                     self.soft.params['min_contamination'])
        stats, bins = '%s.stats' % out, '%s_bins' % out
        self.outputs['dirs'].append(out_dir)
        self.outputs['outs'][group] = [stats, bins]
        io_update(self, i_d=bin_folders, o_d=[stats, bins], key=group)


def get_binners(self, out, binned):
    binners = []
    work_files = '%s/work_files' % out
    for binner, bins in binned.items():
        if self.config.force or not isdir(bins) or not isdir(work_files):
            binners.append(binner)
    return binners


def binning_cmd(self, fastqs, out, contigs, binned):
    binners = get_binners(self, out, binned)
    if binners:
        cmd = 'metawrap binning'
        cmd += ' -o %s' % out
        cmd += ' -a %s' % contigs
        cmd += ' -t %s' % self.soft.params['cpus']
        cmd += ' -l %s' % self.soft.params['min_contig_length']
        cmd += ' --universal'
        for binner in binners:
            cmd += ' --%s' % binner
        cmd += ' %s' % ' '.join(fastqs)
        return cmd


def binning(self):
    for group in self.pools[self.pool]:
        tmp = '$TMPDIR/mtwrp_%s_%s' % (self.pool, group)
        out = '%s/%s/%s' % (self.dir, self.pool, group)
        self.outputs['dirs'].append(out)
        binned = {binner: '%s/%s_bins' % (out, binner)
                  for binner in self.soft.params['binners']}
        # outputs
        bin_dirs = sorted(binned.values()) + ['work_files']
        self.outputs['outs'][group] = bin_dirs
        # inputs
        contigs = self.inputs[self.pool][group][1]
        fastqs = [fastq for sam in self.pools[self.pool][group]
                  for fastq in self.config.fastq[sam]]
        cmd = binning_cmd(self, fastqs, out, contigs, binned)
        if cmd:
            self.outputs['cmds'].setdefault(group, []).append(cmd)
            io_update(self, i_f=([contigs] + fastqs), i_d=tmp,
                      o_d=bin_dirs, key=group)


def metawrap(self) -> None:
    """Create command lines for metaWRAP

    Parameters
    ----------
    self : Commands class instance
        Contains all the attributes needed for binning on the current sample
    """
    # This function splits the name of the software and calls as function
    # the last underscore-separated field (which is in this module)
    caller(self, __name__)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import argparse


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', nargs=1, required=True, help='sam file')
    parser.add_argument('-o', nargs=1, required=True, help='output sam file')
    parser.add_argument('-s', nargs=1, required=True, help='sample name')
    parse = parser.parse_args()
    args = vars(parse)
    return args


def write_formatted(filin, filou, sample):
    with open(filou, 'w') as o, open(filin) as f:
        idx = 0
        for line in f:
            if line[0] == '@':
                continue
            idx += 1
            ls = line.split()
            o.write('%s_%s_%s\t%s\n' % (sample, idx, ls[0], ls[-1]))


if __name__ == '__main__':
    args = get_args()
    filin = args['i'][0]
    filou = args['o'][0]
    sample = args['s'][0]
    write_formatted(filin, filou, sample)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
from os.path import basename, isfile, splitext
from metagenomix._io_utils import io_update


def get_quast_contigs_labels(self, pool, group_samples):
    contigs = []
    labels = []
    for group, fps in sorted(self.inputs[pool].items()):
        contigs.append(fps[1])
        if self.soft.params['label']:
            samples = group_samples[group]
            vals = self.config.meta.loc[samples, self.soft.params['label']]
            labels.append('__'.join(sorted(set(vals))))
    return contigs, labels


def quast(self):
    for pool, group_samples in self.pools.items():
        contigs, labels = get_quast_contigs_labels(self, pool, group_samples)

        out_dir = '%s/%s' % (self.dir, pool)
        if self.soft.params['label']:
            out_dir += '_%s' % self.soft.params['label']

        cmd = 'metaquast.py'
        cmd += ' %s' % ' '.join(contigs)
        if labels:
            cmd += ' --labels %s' % ','.join(labels)
        cmd += ' --output-dir %s' % out_dir
        cmd += ' --min-contig %s' % self.soft.params['min_contig']
        cmd += ' --ambiguity-usage %s' % self.soft.params['ambiguity_usage']
        for boolean in ['circos', 'glimmer', 'rna_finding',
                        'conserved_genes_finding', 'space_efficient']:
            if self.soft.params[boolean]:
                cmd += ' --%s' % boolean.replace('_', '-')
        if self.soft.params['fast']:
            cmd += ' --fast'
        else:
            for boolean in ['no_check', 'no_plots', 'no_html', 'no_icarus',
                            'no_snps', 'no_gc', 'no_sv', 'no_read_stats']:
                if self.soft.params[boolean]:
                    cmd += ' --%s' % boolean.replace('_', '-')

        self.outputs['dirs'].append(out_dir)
        self.outputs['outs'][pool] = out_dir
        self.outputs['cmds'].setdefault(pool, []).append(cmd)
        io_update(self, i_f=contigs, o_d=out_dir, key=pool)


def spades_cmd(fastas: list, params: dict, tmp: str, out: str):
    """Create command lines for spades.

    Parameters
    ----------
    fastas : list
        Input fasta files
    params : dict
        Parameters for Spades
    tmp : str
        Temporary folder for Spades
    out : str
        Output folder for Spades
    """
    cmd = 'spades.py'
    cmd += ' -m %s' % params['mem_num']
    cmd += ' -k %s' % ','.join(map(str, params['k']))
    cmd += ' -t %s' % params['cpus']
    if params['bio']:
        cmd += ' --bio'
    if params['meta']:
        cmd += ' --meta'
    if params['plasmid']:
        cmd += ' --plasmid'
    if params['only_assembler']:
        cmd += ' --only-assembler'
    cmd += ' --tmp-dir %s -o %s' % (tmp, out)
    for fasta in fastas:
        if 'extendedFrags' in fasta:
            cmd += ' --merge %s' % fasta
        elif 'notCombined_1' in fasta:
            cmd += ' -1 %s' % fasta
        elif 'notCombined_2' in fasta:
            cmd += ' -2 %s' % fasta
    return cmd


def spades(self) -> None:
    """Create command lines for spades.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for spades
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for group, fastas in self.inputs[self.pool].items():
        tmp = '$TMPDIR/spades_%s_%s' % (self.pool, group)
        out = '%s/%s/%s' % (self.dir, self.pool, group)
        cmd = 'mkdir -p %s\n' % tmp
        cmd += '%s\n' % spades_cmd(fastas, self.soft.params, tmp, out)
        cmd += 'rm -rf %s\n' % tmp
        before_rr = '%s/before_rr.fasta' % out
        contigs = '%s/contigs.fasta' % out
        first_pe = '%s/first_pe_contigs.fasta' % out
        scaffolds = '%s/scaffolds.fasta' % out
        log = '%s/spades.log' % out
        outs = [before_rr, contigs, first_pe, scaffolds, log]
        if self.config.force or not isfile(contigs):
            self.outputs['cmds'].setdefault(group, []).append(cmd)
            io_update(self, i_f=fastas, i_d=tmp, o_f=outs, o_d=out, key=group)
        self.outputs['outs'][group] = outs


def viralverify(self):
    if self.soft.prev != 'spades':
        sys.exit('[viralVerify] can only be run on assembly output')
    pfam = '%s/Pfam-A.hmm' % self.databases.pfams.get('dir')
    if not self.config.dev and not isfile(pfam):
        sys.exit('[viralVerify] Needs a Pfam .hmm database in database config')
    for group, spades_outs in self.inputs[self.pool].items():
        out = '%s/%s/%s' % (self.dir, self.pool, group)
        cmd = 'viralverify'
        cmd += ' -f %s' % spades_outs[1]
        cmd += ' -o %s' % out
        if self.soft.params['db']:
            cmd += ' --db'
        if self.soft.params['p']:
            cmd += ' --p'
        cmd += ' -t %s' % self.soft.params['cpus']
        cmd += ' -thr %s' % self.soft.params['thr']
        cmd += ' --hmm %s' % pfam
        out_fp = '%s_result_table.csv' % splitext(spades_outs[1])[0]
        if self.config.force or not isfile(out_fp):
            self.outputs['cmds'].setdefault(group, []).append(cmd)
            io_update(self, i_f=spades_outs[1], o_d=out, key=group)
        self.outputs['outs'][group] = out_fp


def plass(self):
    """Create command lines for Plass.

    self : Commands class instance
        .dir : str
            Path to pipeline output folder for spades
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    out = '%s/%s' % (self.dir, self.sam)
    self.outputs['dirs'].append(out)
    inputs = self.inputs[self.sam]
    if self.soft.prev == 'flash':
        inputs = [x.replace('.fasta', '.fastq') for x in inputs if
                  'notCombined_' in basename(x)]
    out_fp = '%s/plass_%sassembly.fasta' % (out, self.soft.params.type)
    self.outputs['outs'].append(out_fp)
    tmp_dir = '$TMPDIR/plass_%s' % self.sam
    cmd = 'plass %sassemble' % self.soft.params.type
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --num-iterations 6'
    cmd += ' --min-seq-id 0.8'
    cmd += ' -c 0.8'
    cmd += ' %s' % ' '.join(sorted(inputs))
    cmd += ' %s' % out_fp
    cmd += ' %s' % tmp_dir
    if self.config.force or not isfile(out_fp):
        cmd = 'mkdir -p %s\n%s\nrm -rf %s' % (tmp_dir, cmd, tmp_dir)
        self.outputs['cmds'].append(cmd)
        io_update(self, i_f=inputs, i_d=tmp_dir, o_f=out_fp)


def get_merges(self) -> list:
    """Get the extension of the file to merge to make the pools.

    Parameters
    ----------
    self : Commands class instance
        .soft.prev
            Software run just before pooling

    Returns
    -------
    merges : list
        Extension of the outputs to merge
    """
    merges = ['']
    if self.soft.prev == 'flash':
        merges = ['extendedFrags', 'notCombined_1', 'notCombined_2']
    return merges


def pool_cmd(self, pool: str, out_dir: str, paths: list,
             fasta: str, group: str) -> None:
    """Write the pooling command and collect the output and io for FLASh.

    Parameters
    ----------
    self : Commands class instance
        .cmds
            Command lines
        .soft.io : dict
            All files to I/O is scratch is used
        .config
            Configurations
    pool : str
        Name of the pool
    out_dir : str
        Path to the output directory
    paths : list
        Path the input fasta files
    fasta : str
        Path to an output fasta file
    group : str
        Name of the sample group within the pool
    """
    if self.config.force or not isfile(fasta):
        for pdx, path in enumerate(paths):
            if pdx:
                cmd = 'cat %s >> %s' % (path, fasta)
            else:
                cmd = 'cat %s > %s' % (path, fasta)
            self.cmds.setdefault(pool, []).append(cmd)
        io_update(self, i_f=paths, o_f=fasta, o_d=out_dir, key=group)


def get_pools(self, pool: str, group: str, sams: list) -> list:
    """Get the output fasta files returned by FLASh and write the commands.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for spades
        .inputs : dict
            Input files
        .cmds
            Command lines
        .soft.dirs : set
            All directories to create
        .soft.io : dict
            All files to I/O is scratch is used
        .config
            Configurations
    pool : str
        Name of the pool
    group : str
        Name of the sample group within the pool
    sams : list
        Samples to pool

    Returns
    -------
    fasta_fps : list
        Paths to the output fasta files
    """
    out_dir = self.dir + '/' + pool
    self.soft.dirs.add(out_dir.replace('${SCRATCH_FOLDER}', ''))
    fasta_fps = []
    for merge in get_merges(self):
        paths = [fp for sam in sams for fp in self.inputs[sam] if merge in fp]
        # only pool if there is min 2 samples being merged
        if len(sams) > 1:
            fasta = '%s/%s.%s.fasta' % (out_dir, group, merge)
            fasta = fasta.replace(' ', '_').replace('..', '.')
            pool_cmd(self, pool, out_dir, paths, fasta, group)
        else:
            if len(paths) > 1:
                raise ValueError('Error in pooling group...')
            fasta = paths[0]
        fasta_fps.append(fasta)
    return fasta_fps


def pooling(self, pool):
    """Create command lines for pooling samples' reads using FLASh.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for spades
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    pool : str
        Name of the pool
    """
    for group, group_pd in self.config.meta.groupby(pool):
        # get the full list of samples and the samples per pooling group
        sams = group_pd.index.tolist()
        self.pools[pool][group] = sams
        # get the outputs for the current group and collect pooling commands
        self.soft.outputs[pool][group] = get_pools(self, pool, group, sams)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import pandas as pd
from os.path import isfile

from prep_songbird._io_utils import (
    get_taxonomy_classifier, get_analysis_folder, parse_g2lineage)
from prep_songbird._cmds import (
    write_seqs_fasta, write_taxonomy_sklearn, run_export, run_import)


def edit_split_taxonomy(
        ranks: dict, split_taxa_pd: pd.DataFrame) -> pd.DataFrame:
    if len(ranks) == split_taxa_pd.shape[1]:
        split_taxa_pd = split_taxa_pd.rename(columns=ranks)
    else:
        alpha = 'ABCDEFGHIJKLMNOPQRST'
        cols = [alpha[x] for x in range(split_taxa_pd.shape[1])]
        split_taxa_pd = pd.DataFrame([
            pd.Series([
                '%s__%s' % (cols[idx], str(x).replace(' ', '_'))
                for idx, x in enumerate(row) if str(x) != 'nan']
            ) for row in split_taxa_pd.values],
            columns=cols,
            index=split_taxa_pd.index.tolist()
        )
    return split_taxa_pd


def get_ranks_from_split_taxonomy(split_taxa_pd, col):
    rank = [x.split('__')[0] for x in split_taxa_pd[col]
            if str(x) not in ['nan', 'None', 'Unassigned']]
    if len(set(rank)) == 1:
        return rank[0]
    else:
        rank = [x.split('_')[0] for x in split_taxa_pd[col]
                if str(x) not in ['nan', 'None', 'Unassigned']]
        if len(set(rank)) == 1:
            return rank[0]
    return ''


def parse_split_taxonomy(split_taxa_pd: pd.DataFrame) -> dict:
    torm = []
    ranks = {}
    # not_collapsable = False
    # parse each column/levels to determine which is
    # (i) to be removed or (ii) labelled with consistent rank (e.g. "k__"))
    for col in split_taxa_pd.columns:
        # if some levels of the split taxonomy are the feature IDs themselves:
        # remove these levels (as it would be a copy of the non-collapsed table)
        col_features = split_taxa_pd[col].tolist()
        if split_taxa_pd.index.tolist() == col_features:
            torm.append(col)
        else:
            rank = get_ranks_from_split_taxonomy(split_taxa_pd, col)
            if rank and rank != col:
                ranks[col] = rank
    # remove columns to be removed
    if torm:
        split_taxa_pd.drop(columns=torm, inplace=True)
    return ranks


def get_split_taxonomy(taxa, taxo_sep=';'):
    split_lens = set()
    split_taxa = []
    for tdx, taxon in enumerate(taxa):
        if str(taxon) == 'nan':
            split_lens.add(1)
            split_taxa.append(pd.Series(['Unassigned']))
        else:
            taxon_split = [x.strip() for x in str(taxon).split(taxo_sep)
                           if len(x.strip()) and not x.startswith('x__')]
            split_lens.add(len(taxon_split))
            split_taxa.append(pd.Series(taxon_split))
    # if the parsed and split taxonomies have  very variable number of fields
    # or very long split results  it is terminated here as a taxonomy that
    # does not make sense
    if len(split_lens) > 15 or max(split_lens) > 15:
        return pd.DataFrame([[x] for x in taxa], columns=['not_really_taxon'])
    # build a dataframe from the spit taxonomy
    split_taxa_pd = pd.DataFrame(split_taxa)
    # add a column label
    ALPHA = 'ABCDEFGHIJKLMNOPQRST'
    split_taxa_pd.columns = ['Taxolevel_%s' % (ALPHA[idx])
                             for idx in range(split_taxa_pd.shape[1])]
    return split_taxa_pd


def get_tax_tables(tax_fp: str) -> tuple:
    # read taxonomy with features as index, format and collect features IDs list
    tax_pd = pd.read_csv(tax_fp, header=0, sep='\t', dtype=str)
    tax_pd.rename(columns={tax_pd.columns[0]: 'Feature ID'}, inplace=True)
    # perform taxonomic split on the Taxon list and give the Feature as index
    split_taxa_pd = get_split_taxonomy(tax_pd.Taxon.tolist())
    split_taxa_pd.index = tax_pd['Feature ID'].tolist()
    return tax_pd, split_taxa_pd


def get_taxa_edit(taxo):
    taxo_edit = []
    for tax in taxo:
        if str(tax) == 'nan':
            taxo_edit.append(tax)
        elif not tax.strip('_'):
            taxo_edit.append(tax)
        else:
            taxo_edit.append(tax.replace(',', '_'))
    return taxo_edit


def get_edit_taxonomy_command(data):
    cmd = ''
    out_pd = pd.read_csv(data.tax[2], dtype=str, sep='\t')
    taxo = out_pd['Taxon'].tolist()
    taxo_edit = get_taxa_edit(taxo)
    if taxo != taxo_edit:
        out_pd['Taxon'] = taxo_edit
        out_pd.to_csv(data.tax[2], index=False, sep='\t')
        cmd = run_import(
            data.tax[2], data.tax[1], 'FeatureData[Taxonomy]')
    return cmd


def run_taxonomy_wol(force: bool, tsv_pd: pd.DataFrame, out_qza: str,
                     out_tsv: str, cur_datasets_features: dict) -> str:
    cmd = ''
    if force or not isfile(out_qza):
        g2lineage = parse_g2lineage()
        rev_cur_datasets_features = dict(
            (y, x) for x, y in cur_datasets_features.items())
        if not isfile(out_tsv):
            with open(out_tsv, 'w') as o:
                o.write('Feature ID\tTaxon\n')
                for feat in tsv_pd.index:
                    if rev_cur_datasets_features[feat] in g2lineage:
                        o.write('%s\t%s\n' % (
                            feat, g2lineage[rev_cur_datasets_features[feat]]))
                    else:
                        o.write('%s\t%s\n' % (feat, feat.replace('|', '; ')))
        cmd = run_import(out_tsv, out_qza, 'FeatureData[Taxonomy]')
    return cmd


def run_taxonomy_amplicon(
        dat: str, i_datasets_folder: str, force: bool, tsv_pd: pd.DataFrame,
        out_qza: str, out_tsv: str, i_classifier: str) -> str:
    cmd = ''
    if isfile(out_tsv) and not isfile(out_qza):
        cmd += run_import(out_tsv, out_qza, 'FeatureData[Taxonomy]')
    else:
        ref_classifier_qza = get_taxonomy_classifier(i_classifier)
        odir_seqs = get_analysis_folder(i_datasets_folder, 'seqs/%s' % dat)
        out_fp_seqs_rad = '%s/seq_%s' % (odir_seqs, dat)
        out_fp_seqs_fasta = '%s.fasta' % out_fp_seqs_rad
        out_fp_seqs_qza = '%s.qza' % out_fp_seqs_rad
        if force or not isfile(out_fp_seqs_qza):
            cmd += write_seqs_fasta(out_fp_seqs_fasta, out_fp_seqs_qza, tsv_pd)
        if force or not isfile(out_qza):
            cmd += write_taxonomy_sklearn(out_qza, out_fp_seqs_qza,
                                          ref_classifier_qza)
            cmd += run_export(out_qza, out_tsv, '')
    return cmd


def run_taxonomy_others(force: bool, tsv_pd: pd.DataFrame,
                        out_qza: str, out_tsv: str) -> str:
    cmd = ''
    if force or not isfile(out_qza):
        if not isfile(out_tsv):
            with open(out_tsv, 'w') as o:
                o.write('Feature ID\tTaxon\n')
                for feat in tsv_pd.index:
                    o.write('%s\t%s\n' % (feat, feat))
        cmd = run_import(out_tsv, out_qza, 'FeatureData[Taxonomy]')
    return cmd


def get_taxonomy_command(dat, config, data):
    cmd = ''
    if data.tax[0] == 'wol':
        cmd = run_taxonomy_wol(
            config.force, data.data[0], data.tax[1],
            data.tax[2], data.features)
    elif data.tax[0] in ['amplicon', 'sklearn']:
        if config.i_classifier:
            cmd = run_taxonomy_amplicon(
                dat, config.i_datasets_folder, config.force, data.data[0],
                data.tax[1], data.tax[2], config.i_classifier)
        else:
            print('No classifier passed for 16S data\nExiting...')
    else:
        cmd = run_taxonomy_others(
            config.force, data.data[0], data.tax[1], data.tax[2])
    return cmd


